###################MicrobiomeAnalyst R code (Bckend)#################
.on.public.web <- FALSE;
.set.mbSetObj <- function(mbSetObj=NA){
  if(.on.public.web){
    mbSet <<- mbSetObj;
    return (1);
  }
  return(mbSetObj);
}
.on.public.web <- FALSE;

.get.mbSetObj <- function(mbSetObj=NA){
  if(.on.public.web){
    return(mbSet);
  }else{
    return(mbSetObj);
  }
}
Init.mbSetObj <- function(){
  
  dataSet <- list();
  analSet <- list();
  imgSet <- list();
  
  mbSetObj <- list();
  mbSetObj$dataSet <- dataSet;
  mbSetObj$analSet <- analSet;
  mbSetObj$imgSet <- imgSet;
  mbSetObj$is.ASV <- FALSE;
  mbSetObj$poor.replicate <- FALSE; # flag to show if all unique values
  mbSetObj$tree.uploaded <- FALSE;
  mbSetObj$cmdSet <- vector(mode="character"); # store R command
  
  # set global variables
  current.msg <<- "";
  current.selected.tax <<- "NA";
  enrich.type <<- "hyper";
  
  load_cairo();
  load_ggplot();
  BiocParallel::register(BiocParallel::SerialParam());
  corr.net.count <<- 0
  
  # preload some general package
  Cairo::CairoFonts("Arial:style=Regular","Arial:style=Bold","Arial:style=Italic","Helvetica","Symbol")
  print("Init MicrobiomeAnalyst!");
  return(.set.mbSetObj(mbSetObj))
}
load_cairo <- function(){
  suppressMessages(library(Cairo))
}
SetModuleType <- function(mbSetObj, nm){
  mbSetObj <- .get.mbSetObj(mbSetObj);
  mbSetObj$module.type <- nm;
  return(.set.mbSetObj(mbSetObj));
}
load_phyloseq <- function(){
  suppressMessages(library(phyloseq))
}
load_ggplot <- function(){
  suppressMessages(library(ggplot2))
}
.readDataTable <- function(fileName){
  dat <- try(data.table::fread(fileName, header=TRUE, check.names=FALSE, blank.lines.skip=TRUE, data.table=FALSE));
  if(class(dat) == "try-error" || any(dim(dat) == 0)){
    print("Using slower file reader ...");
    formatStr <- substr(fileName, nchar(fileName)-2, nchar(fileName))
    if(formatStr == "txt"){
      dat <- try(read.table(fileName, header=TRUE, comment.char = "", check.names=F, as.is=T));
    }else{ # note, read.csv is more than read.table with sep=","
      dat <- try(read.csv(fileName, header=TRUE, comment.char = "", check.names=F, as.is=T));
    }
  }
  return(dat);
}
ReadSampleTable<- function(mbSetObj, dataName) {
  
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  load_phyloseq();
  
  msg <- NULL;
  mydata <- .readDataTable(dataName);
  
  if(any(is.na(mydata)) || class(mydata) == "try-error"){
    current.msg <<- "Failed to read in the metadata! Please make sure that the metadata file is in the right format and does not have empty cells or NA.";
    return(0);
  }
  
  # look for #NAME, store in a list
  sam.nm <- substr(colnames(mydata[1]),1,5);
  sam.nm <- tolower(sam.nm);
  sam.inx <- grep("#name",sam.nm);
  
  if(length(sam.inx) > 0){
    smpl_nm<-mydata[,1];
    smpl_var<-colnames(mydata[-1]);
  }else{
    current.msg <<- "Please make sure you have the label #NAME in your sample data file!";
    return(0);
  }
  
  # converting to character matrix as duplicate row names not allowed in data frame.
  mydata <- as.matrix(mydata[,-1]);
  rownames(mydata) <- smpl_nm;
  colnames(mydata) <- smpl_var;
  
  # empty cell or NA cannot be tolerated in metadata
  na.inx  <- is.na(mydata);
  na.msg <- NULL;
  
  if(sum(na.inx) > 0){
    mydata[na.inx] <- "Unknown";
    na.msg <- paste("A total of", sum(na.inx), "empty or NA values were replaced by 'Unknown'.");
  }
  
  # as most functions are for discrete groups (not continuouse values
  # require at least one column contains discrete factors with at least two replicates 
  my.meta <- data.frame(mydata);
  mbSetObj$dataSet$group_names <- colnames(my.meta)
  
  disc.inx <- GetDiscreteInx(my.meta);
  if(sum(disc.inx) == 0){ # all class labels are unique! 
    na.msg <- c(na.msg, "It seems that all your meta data values are unique! MicrobiomeAnalyst requires some biological replicates for robust analysis");
    mbSetObj$poor.replicate <- TRUE;
    mbSetObj$dataSet$sample_data <- my.meta
  }else{
    mbSetObj$dataSet$meta_info$disc.inx <- disc.inx;
    mbSetObj$dataSet$sample_data <- my.meta[,disc.inx, drop=FALSE];
    
    cont.inx <- GetNumbericalInx(my.meta);
    cont.inx <- !disc.inx & cont.inx; # discrete is first
    mbSetObj$dataSet$meta_info$cont.inx <- cont.inx;
    
    if(sum(cont.inx)>0){
      # make sure the discrete data is on the left side
      mbSetObj$dataSet$sample_data <- cbind(mbSetObj$dataSet$sample_data, my.meta[,cont.inx, drop=FALSE]);
    }
  }
  
  current.msg <<- paste(na.msg, "The sample data contains a total of ", nrow(mydata), "samples and  ", ncol(mydata), " sample variables.", collapse=" ");
  mbSetObj$dataSet$smpl.msg <- current.msg;
  
  return(.set.mbSetObj(mbSetObj));
}
GetDiscreteInx <- function(my.dat, min.rep=2){
  good.inx <- apply(my.dat, 2, function(x){
    good1.inx <- length(x) > length(unique(x));
    good2.inx <- min(table(x)) >= min.rep;
    return (good1.inx & good2.inx);
  });
  return(good.inx);
}
GetNumbericalInx <- function(my.dat){
  good.inx <- apply(my.dat, 2, function(x){
    return(all(!is.na(as.numeric(as.character(x))))); 
  });
  return(good.inx);
}
Read16STaxaTable <- function(mbSetObj, dataName) {
  
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  msg <- NULL;
  mydata <- .readDataTable(dataName);
  
  if(is.null(mydata) || class(mydata) == "try-error"){
    current.msg <<- "Failed to read in the taxonomic data! Please make sure the data is in the right format.";
    return(0);
  }
  
  # look for #TAXONOMY, store in a list
  sam.nm <- substr(colnames(mydata[1]),1,9);
  sam.nm <- tolower(sam.nm);
  sam.inx <- grep("^#taxonomy",sam.nm);
  
  if(length(sam.inx) > 0){
    tax_nm <- mydata[,1];
    tax_rank <- colnames(mydata[-1]);
  }else{
    current.msg <<- "No labels #TAXONOMY found in your data!";
    return(0);
  }
  
  # converting to character matrix as duplicate row names not allowed in data frame.
  mydata <- as.matrix(mydata[,-1]);
  rownames(mydata) <- tax_nm;
  colnames(mydata) <- tax_rank;
  current.msg <<- paste("Taxonomy file has total of ",nrow(mydata),"features and",ncol(mydata), "taxonomic rank. Additional features which are not present in abundance table has been removed if present.");
  
  mbSetObj$dataSet$taxa_table <- mydata;
  
  return(.set.mbSetObj(mbSetObj));
}
Read16STabData <- function(mbSetObj, dataName) {
  
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  msg <- NULL;
  mydata <- .readDataTable(dataName);
  
  if(any(is.na(mydata)) || class(mydata) == "try-error"){
    current.msg <<- "Failed to read in the OTU abundance data! Please make sure the data is in the right format and do not have empty cells or NA.";
    return(0);
  }
  
  
  # note dateSet$sample_data already set
  # getting NAME label
  sam.nm <- substr(colnames(mydata[1]),1,5);
  sam.nm <- tolower(sam.nm);
  sam.inx <- grep("^#name",sam.nm);
  
  if(length(sam.inx) == 0){
    current.msg <<- "No labels #NAME found in your data!";
    return(0);
  }   
  smpl_nm <- colnames(mydata[-1]);
  
  
  # remove extra comments if any
  dat.nms <- mydata[,1];
  comment.inx <- grep("^#",dat.nms);
  
  if(sum(comment.inx) > 0){
    mydata < -mydata[-comment.inx, ];
    msg <- c(msg, paste("A total of", sum(comment.inx), "comment rows were removed from OTU table"));
    dat.nms <- mydata[,1];
  }
  
  mydata <- data.matrix(mydata[,-1]);
  rownames(mydata) <- dat.nms;
  
  # empty cell or NA cannot be tolerated in metadata
  na.inx  <- is.na(mydata);
  
  if(sum(na.inx) > 0){
    mydata[na.inx] <- "Unknown";
    msg <- c(msg, paste("A total of", sum(na.inx), "empty or NA values were replaced by zeros."));
  }
  
  msg <<- paste(msg, collapse="; ");
  
  mbSetObj$dataSet$name <- basename(dataName);
  mbSetObj$dataSet$smpl_nm <- smpl_nm;
  mbSetObj$dataSet$data.orig <- data.matrix(mydata);
  
  mbSetObj$dataSet$read <- TRUE
  return(.set.mbSetObj(mbSetObj))
}
Read16SAbundData <- function(mbSetObj, dataName, format, taxa_type, ismetafile) {
  
  load_phyloseq();
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  if(format=="text"){
    if(.on.public.web){
      if(Read16STabData(mbSetObj, dataName)==0){
        return(0);
      }
    }else{
      # offline
      mbSetObj <- Read16STabData(mbSetObj, dataName)
      if(!mbSetObj$dataSet$read){
        return(0);
      }
    }
    
  }else if(format=="biom"){
    if(.on.public.web){
      if(Read16SBiomData(mbSetObj, dataName, taxa_type, ismetafile)==0){
        return(0);
      } 
    }else{
      # offline
      mbSetObj <- Read16SBiomData(mbSetObj, dataName, taxa_type, ismetafile);
      if(!mbSetObj$dataSet$read){
        return(0);
      }
    }
  }
  
  #need to refetch mbSetObj
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  data.orig <- mbSetObj$dataSet$data.orig;
  
  if(mode(data.orig)=="character"){
    current.msg <<- paste("Errors in parsing your data as numerical - possible reason: comma as decimal separator?");
    return(0);
  }
  
  #storing whole taxonomy label(used for PICRUST & Tax4Fun; reference data mapping)
  mbSetObj$dataSet$comp_taxnm <- rownames(data.orig);
  current.msg <<- paste("A total of",ncol(data.orig) ,"samples and ", nrow(data.orig), "features or taxa are present.");
  mbSetObj$dataSet$read.msg <- current.msg;
  mbSetObj$dataSet$data.type <- format;
  mbSetObj$dataSet$taxa.type <- taxa_type;
  
  return(.set.mbSetObj(mbSetObj));
  
}

SanityCheckData <- function(mbSetObj, filetype){
  
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  feat.sums <- apply(mbSetObj$dataSet$data.orig, 1, function(x){sum(x>0, na.rm=T)});
  gd.inx <- feat.sums > 1; # occur in at least 2 samples
  
  if(length(which(gd.inx=="TRUE"))==0){
    current.msg <<-"Reads occur in only one sample.  All these are considered as artifacts and have been removed from data. No data left after such processing.";
    return(0);
  }
  
  data.proc <- mbSetObj$dataSet$data.orig[gd.inx, ]; 
  # filtering the constant features here
  # check for columns with all constant (var=0)
  varCol <- apply(data.proc, 1, var, na.rm=T);
  constCol <- varCol == 0 | is.na(varCol);
  
  # making copy of data.proc and proc.phyobj(phyloseq)
  data.proc <- data.proc[!constCol, ];
  
  if(length(data.proc)==0){
    current.msg <<-"All features are found to be constant and have been removed from data. No data left after such processing.";
    return(0);
  }
  
  saveRDS(data.proc, file="data.proc.orig"); # save an copy
  saveRDS(data.proc, file="data.prefilt"); # save an copy
  
  # now get stats
  taxa_no <- nrow(mbSetObj$dataSet$data.orig);
  
  # from abundance table
  sample_no <- ncol(mbSetObj$dataSet$data.orig);
  
  if(filetype=="biom"||filetype=="mothur"){
    samplemeta_no <- nrow(mbSetObj$dataSet$sample_data);
  }else{
    samplemeta_no <- sample_no;
  }
  
  if(mbSetObj$module.type!="ppd"){
    mbSetObj$dataSet$sample_data <- mbSetObj$dataSet$sample_data[sapply(mbSetObj$dataSet$sample_data, function(col) length(unique(col))) > 1];
  }
  
  if(ncol(mbSetObj$dataSet$sample_data)==0){
    current.msg <<-"No sample variable have more than one group. Please provide variables with at least two groups.";
    return(0);
  }
  
  #converting all sample variables to factor type
  character_vars <- sapply(mbSetObj$dataSet$sample_data, is.character);
  
  if(any(character_vars)=="TRUE"){
    mbSetObj$dataSet$sample_data[, character_vars] <- sapply(mbSetObj$dataSet$sample_data[, character_vars], as.factor);
  }
  
  num_vars <- sapply(mbSetObj$dataSet$sample_data, is.numeric);
  
  if(any(num_vars)=="TRUE"){
    mbSetObj$dataSet$sample_data[, num_vars] <- sapply(mbSetObj$dataSet$sample_data[, num_vars],as.factor);
  }
  
  if(file.exists("tree.RDS")){
    tree_exist <- 1
  } else {
    tree_exist <- 0
  }
  
  if(identical(sort(row.names(mbSetObj$dataSet$sample_data)),
               sort(colnames(data.proc)))){
    samname_same <- 1;
  } else {
    samname_same <- 0;
  }
  
  sample_no_in_outfile <- ncol(data.proc);
  samname_same_number <- sum(row.names(mbSetObj$dataSet$sample_data) %in% colnames(data.proc));
  
  # now store data.orig to RDS
  saveRDS(mbSetObj$dataSet$data.orig, file="data.orig");
  saveRDS(data.proc, file="data.proc");
  saveRDS(mbSetObj$dataSet$sample_data, file = "data.sample_data")
  mbSetObj$dataSet$data.orig <- NULL;
  mbSetObj$dataSet$tree <- tree_exist
  
  vari_no <- ncol(mbSetObj$dataSet$sample_data);
  disc_no <- sum(mbSetObj$dataSet$meta_info$disc.inx);
  cont_no <- sum(mbSetObj$dataSet$meta_info$cont.inx);
  smpl.sums <- apply(data.proc, 2, sum);
  tot_size <- sum(smpl.sums);
  smin <- min(smpl.sums)
  smean <- mean(smpl.sums)
  smax <- max(smpl.sums);
  gd_feat <- nrow(data.proc);
  
  if(.on.public.web){
    .set.mbSetObj(mbSetObj)
    return(c(1,taxa_no,sample_no,vari_no, smin,smean,smax,gd_feat,samplemeta_no,tot_size, tree_exist, samname_same, samname_same_number, sample_no_in_outfile, disc_no, cont_no));
  }else{
    print("Sanity check passed!")
    return(.set.mbSetObj(mbSetObj))
  }
}
GetExtendRange<-function(vec, unit=10){
  var.max <- max(vec, na.rm=T);
  var.min <- min(vec, na.rm=T);
  exts <- (var.max - var.min)/unit;
  c(var.min-exts, var.max+exts);
}
PlotLibSizeView <- function(mbSetObj, imgName, format="png", dpi=72){
  
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  data.proc <- readRDS("data.proc");
  data_bef <- data.matrix(data.proc);
  smpl.sums <- colSums(data_bef);
  names(smpl.sums) <- colnames(data_bef);
  smpl.sums <- sort(smpl.sums);
  smpl.sums <- rev(smpl.sums);
  vip.nms <- names(smpl.sums);
  names(smpl.sums) <- NULL;
  vip.nms <- substr(vip.nms, 1, 16);
  
  myH <- ncol(data_bef)*25 + 50;
  imgName = paste(imgName,".", format, sep="");
  mbSetObj$imgSet$lib.size<-imgName;
  Cairo::Cairo(file=imgName, width=580, height=myH, type=format, bg="white",dpi=dpi);
  xlim.ext <- GetExtendRange(smpl.sums, 10);
  par(mar=c(4,7,2,2));
  dotchart(smpl.sums, col="forestgreen", xlim=xlim.ext, pch=19, xlab="Read Counts", main="Library Size Overview");
  mtext(side=2, at=1:length(vip.nms), vip.nms, las=2, line=1)
  text(x=smpl.sums,y=1:length(smpl.sums),labels= round(smpl.sums), col="blue", pos=4, xpd=T);
  dev.off();
  
  return(.set.mbSetObj(mbSetObj))
}
# Load splitstackshape
load_splitstackshape <- function(){
  suppressMessages(library(splitstackshape))
}
CreatePhyloseqObj<-function(mbSetObj, type, taxa_type, taxalabel){
  
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  load_phyloseq();
  load_splitstackshape();
  
  data.proc <- readRDS("data.proc");
  
  # do some sanity check here on sample and feature names
  smpl.nms <- colnames(data.proc);
  taxa.nms <- row.names(data.proc);
  
  # check for uniqueness of dimension name
  if(length(unique(smpl.nms))!=length(smpl.nms)){
    dup.nms <- paste(smpl.nms[duplicated(smpl.nms)], collapse="; ");
    current.msg <<- paste(c("Duplicate sample names are not allowed:"), dup.nms, collapse=" ");
    return(0);
  }
  
  # check for uniqueness of taxon names for metagenomic data only
  if(mbSetObj$module.type == "sdp"){
    if(length(unique(taxa.nms))!=length(taxa.nms)){
      dup.tax.nms <- paste(taxa.nms[duplicated(taxa.nms)], collapse="; ");
      current.msg <<- paste(c("Duplicate taxon names are not allowed:"), dup.tax.nms, collapse=" ");
      return(0);
    }
  }else{
    # for ASV data, replace sequences with ASV1, ASV2 ....
    # check is avegage name length is > 75
    nm.ln <- sum(nchar(taxa.nms)/length(taxa.nms));
    if(nm.ln > 75){
      mbSetObj$is.ASV <-TRUE;
    }
  }
  
  # now check for special characters in the data labels
  if(sum(is.na(iconv(smpl.nms)))>0){
    na.inx <- is.na(iconv(smpl.nms));
    nms <- paste(smpl.nms[na.inx], collapse="; ");
    AddErrMsg(paste("No special letters (i.e. Latin, Greek) are allowed in sample names:", nms, collapse=" "));
    return(0);
  }
  
  var.nms <- rownames(data.proc);
  
  if(sum(is.na(iconv(var.nms)))>0){
    na.inx <- is.na(iconv(var.nms));
    nms <- paste(var.nms[na.inx], collapse="; ");
    AddErrMsg(paste("No special letters (i.e. Latin, Greek) are allowed in feature or taxa names:", nms, collapse=" "));
    return(0);
  }
  
  #standard name to be used
  classi.lvl<- c("Kingdom", "Phylum", "Class", "Order", "Family", "Genus", "Species");
  
  if(mbSetObj$module.type == "mdp" | mbSetObj$module.type == "ppd"){
    if(type=="text"){
      # prepare data for phyloseq visualization.
      # if features names are present in specific taxonomy format (greengenes or silva).
      if(taxalabel=="T"){ #taxa labels present and need to be parsed in OTU table
        feat_nm <- rownames(data.proc);
        mbSetObj$dataSet$feat_nm <- feat_nm;
        
        if(taxa_type=="SILVA"){
          
          load_splitstackshape();
          
          feat_nm<-data.frame(mbSetObj$dataSet$feat_nm);
          names(feat_nm)<-"Rank";
          taxonomy<-splitstackshape::cSplit(feat_nm,"Rank",";");
          taxmat= data.frame(matrix(NA, ncol = 7, nrow = nrow(taxonomy)));
          colnames(taxmat) <- classi.lvl;
          taxmat[,1:ncol(taxonomy)]<-taxonomy;
          taxmat<-taxmat[colSums(!is.na(taxmat))>0];
          taxmat<-as.matrix(taxmat);
          rownames(taxmat)<-c(1:nrow(taxmat));
          #phyloseq taxonomy object
          taxa_table <- tax_table(taxmat);
          taxa_names(taxa_table)<-rownames(taxmat);
        }else if(taxa_type=="Greengenes"||taxa_type=="QIIME"){
          a<-lapply(feat_nm, parse_taxonomy_qiime); # note, this functions will remove empty strings, replace with NA!?
          taxa_table<-build_tax_table(a);
          
          if(taxa_type=="Greengenes"){
            #additional columns are added(starting with Rank1 etc; has to be removed)
            indx<-grep("^Rank",colnames(taxa_table));
            # need to test if still columns left (could be that users specified wrong tax format)
            if(ncol(taxa_table) == length(indx)){
              current.msg <<-"Error with parsing Greengenes Taxonomy Labels. Please make sure the taxonomy label was specified correctly during data upload.";
              return(0);
            }
            taxa_table<-taxa_table[,-indx];
            #if all taxa ranks of an taxa are NA with pase_taxonomy_qiime function #covert to Not_assigned
            ind<-which(apply(taxa_table,1, function(x)all(is.na(x)))==TRUE);
            taxa_table[ind,]<-rep("Not_Assigned",ncol(taxa_table));
          }
          taxa_names(taxa_table)<-c(1:nrow(taxa_table));
        }else if(taxa_type=="GreengenesID"||taxa_type=="Others/Not_specific"){
          
          # need to parse Taxonomy still!
          load_splitstackshape();               
          
          feat_nm<-data.frame(mbSetObj$dataSet$feat_nm);
          names(feat_nm)<-"Rank";
          taxonomy<-splitstackshape::cSplit(feat_nm,"Rank",";");
          taxmat= data.frame(matrix(NA, ncol = 7, nrow = nrow(taxonomy)));
          colnames(taxmat) <- classi.lvl;
          taxmat[,1:ncol(taxonomy)]<-taxonomy;
          taxmat<-taxmat[colSums(!is.na(taxmat))>0];
          taxmat<-as.matrix(taxmat);
          rownames(taxmat)<-c(1:nrow(taxmat));
          #phyloseq taxonomy object
          taxa_table <- tax_table(taxmat);
          taxa_names(taxa_table)<-rownames(taxmat);
        }
        
        # making unique id for each OTU consist of lowest taxonomy level present followed by row number
        tmat <- as(tax_table(taxa_table), "matrix")
        rowLn <- nrow(tmat);
        colLn <- ncol(tmat);
        num.mat <- matrix(0, nrow=rowLn, ncol=colLn);
        
        for(i in 1:colLn){
          num.mat[,i] <- rep(i, rowLn);
        }
        
        na.inx <- is.na(tmat);
        num.mat[na.inx] <- 0;
        maxInx <- apply(num.mat, 1, max);
        gd.coords <- cbind(1:rowLn, maxInx);
        gd.nms <- tmat[gd.coords];
        mynames <- make.unique(gd.nms, sep="_");
        taxa_names(taxa_table)<-mynames;
        feat_nm<-NULL;
        mbSetObj$dataSet$taxa_table<-taxa_table;
      }else{
        #sanity check: features name of taxonomy table should match feature names in OTU abundance table,
        # Only features in filtered OTU tables are selected additional are removed.
        taxa_table <- mbSetObj$dataSet$taxa_table;
        indx<-match(rownames(data.proc), rownames(taxa_table));
        
        if(anyNA(indx)){
          na.nms <- rownames(data.proc)[is.na(indx)];
          current.msg <<- paste("The following names cannot be found in your taxanomy table:", paste(na.nms, collapse="; "));
          return(0);
        }
        
        # let's address ASV sequence ID here
        if(mbSetObj$is.ASV){
          # do the conversion and save 
          new.nms <- paste("ASV", 1:length(taxa.nms), sep="_");
          master.nms <- cbind("NewID"=new.nms, "OriginalID"=taxa.nms);
          write.csv(master.nms, file="ASV_ID_mapping.csv");
          
          # update the taxa and sample
          names(new.nms) <- taxa.nms;
          rownames(data.proc) <- new.nms[rownames(data.proc)];
          rownames(taxa_table) <- new.nms[rownames(taxa_table)];
          
          # update tree file if uploaded
          if(mbSetObj$tree.uploaded){
            pg_tree <- readRDS("tree.RDS");
            pg_tree$tip.label <- new.nms[pg_tree$tip.label];
            saveRDS(pg_tree, "tree.RDS");
          }
        }
        
        nm<-colnames(taxa_table);
        taxa_table<-as.matrix(taxa_table[indx,]);
        colnames(taxa_table)<-nm;
        #converting taxonomy file(data frame) to phyloseq taxonomy object;keeping the taxonomy names as it is.
        mbSetObj$dataSet$taxa_table<-tax_table(taxa_table);
        taxa_names(mbSetObj$dataSet$taxa_table)<-rownames(taxa_table);
      }
      
      # creating phyloseq object
      data.proc<-apply(data.proc,2,as.integer);
      data.proc<-otu_table(data.proc,taxa_are_rows =TRUE);
      taxa_names(data.proc)<-taxa_names(mbSetObj$dataSet$taxa_table);
      # removing constant column and making standard names
      ind<-apply(mbSetObj$dataSet$taxa_table[,1], 2, function(x) which(x %in% c("Root", "root")));
      
      if(length(ind)>0){
        mbSetObj$dataSet$taxa_table<-mbSetObj$dataSet$taxa_table[,-1];
        #dataSet$taxa_table<-dataSet$taxa_table #newnew
      }
      
      # removing first KINGDOM OR DOMAIN column
      indx<-apply(mbSetObj$dataSet$taxa_table[,1], 2, function(x) which(x %in% c("k__Fungi", "Fungi", "k__Viruses", "Viruses",
                                                                                 "Bacteria", "Archaea", "k__Bacteria", "k__Archea", 
                                                                                 "D_0__Bacteria", "D_0__Archea")));
      
      if(length(indx)>0){
        mbSetObj$dataSet$taxa_table<-mbSetObj$dataSet$taxa_table[,-1];
      }
      
      classi.lvl<- c("Phylum", "Class", "Order", "Family", "Genus", "Species");
      colnames(mbSetObj$dataSet$taxa_table)<-classi.lvl[1:ncol(mbSetObj$dataSet$taxa_table)]; 
      #sanity check: no of sample of both abundance and metadata should match.
      indx<-match(colnames(data.proc), rownames(mbSetObj$dataSet$sample_data));
      
      if(all(is.na(indx))){
        AddErrMsg("Please make sure that sample names and their number are same in metadata and OTU abundance files.");
        return(0);
      }else{
        mbSetObj$dataSet$sample_data<-mbSetObj$dataSet$sample_data[indx, ,drop=FALSE];
      }
      
      mbSetObj$dataSet$sample_data<-sample_data(mbSetObj$dataSet$sample_data, errorIfNULL=TRUE);
      #cleaning up the names#deleting [] if present; and substituting (space,.,/ with underscore(_))
      mbSetObj$dataSet$taxa_table<- gsub("[[:space:]./_-]", "_",mbSetObj$dataSet$taxa_table);
      mbSetObj$dataSet$taxa_table<- gsub("\\[|\\]","",mbSetObj$dataSet$taxa_table);
      
      #sometimes after removal of such special characters rownames beacame non unique; so make it unique
      mynames1<- gsub("[[:space:]./_-]", "_",taxa_names(mbSetObj$dataSet$taxa_table));
      mynames1<-gsub("\\[|\\]","",mynames1);
      
      if(length(unique(mynames1))< length(taxa_names(mbSetObj$dataSet$taxa_table))){
        mynames1 <- make.unique(mynames1, sep="_");
      }
      
      #after cleaning, names might now be NA
      missing.inx <- which(is.na(mynames1))
      missing_vector <- paste0("New_OTU", seq_along(missing.inx))
      mynames1[missing.inx] <- missing_vector
      
    } else if(type == "biom"||type == "mothur"){
      # creating phyloseq object
      feat_nm<-rownames(data.proc);
      data.proc<-apply(data.proc,2,as.integer);
      data.proc<-otu_table(data.proc,taxa_are_rows =TRUE);
      taxa_names(data.proc)<-feat_nm;
      #sanity check: features name of taxonomy table should match feature names in OTU abundance table,Only features in filtered OTU tables are selected additional are removed.
      indx<-match(taxa_names(data.proc),taxa_names(mbSetObj$dataSet$taxa_table));
      
      if(all(is.na(indx))){
        AddErrMsg("Please make sure that features name of the taxonomy table match the feature names in the OTU abundance table.");
        return(0);
      }
      
      mbSetObj$dataSet$taxa_table<-mbSetObj$dataSet$taxa_table[indx,];
      
      # removing constant column (root) if present otherwise just kingdom column from taxonomy table
      ind<-apply(mbSetObj$dataSet$taxa_table[,1], 2, function(x) which(x %in% c("Root", "root")));
      
      if(length(ind)>0){
        mbSetObj$dataSet$taxa_table<-mbSetObj$dataSet$taxa_table[,-1];
        #dataSet$taxa_table<-dataSet$taxa_table #newnew
      }
      
      # removing first KINGDOM OR DOMAIN column
      indx<-apply(mbSetObj$dataSet$taxa_table[,1], 2, function(x) which(x %in% c("k__Fungi", "Fungi", "k__Viruses", "Viruses", 
                                                                                 "Bacteria","Archaea","k__Bacteria","k__Archea", 
                                                                                 "D_0__Bacteria","D_0__Archea")));
      
      if(length(indx)>0){
        mbSetObj$dataSet$taxa_table<-mbSetObj$dataSet$taxa_table[,-1];
      }
      
      classi.lvl<- c("Phylum", "Class", "Order", "Family", "Genus", "Species"); 
      colnames(mbSetObj$dataSet$taxa_table)<-classi.lvl[1:ncol(mbSetObj$dataSet$taxa_table)];
      #sanity check: no of sample of both abundance and metadata should match.
      indx<-match(colnames(data.proc), rownames(mbSetObj$dataSet$sample_data));
      
      if(all(is.na(indx))){
        current.msg <<-"Please make sure that sample names and their number are the same in metadata and OTU abundance files.";
        return(0);
      }else{
        mbSetObj$dataSet$sample_data<-mbSetObj$dataSet$sample_data[indx, ,drop=FALSE];
      }
      
      #cleaning up the names#deleting [] if present; and substituting (space,.,/ with underscore(_))
      mbSetObj$dataSet$taxa_table <- gsub("[[:space:]./_-]", "_",mbSetObj$dataSet$taxa_table);
      mbSetObj$dataSet$taxa_table<- gsub("\\[|\\]","",mbSetObj$dataSet$taxa_table);
      
      #sometimes after removal of such special characters rownames beacame non unique; so make it unique
      mynames1 <- gsub("[[:space:]./_-]", "_",taxa_names(mbSetObj$dataSet$taxa_table));
      mynames1<-gsub("\\[|\\]","",mynames1);
      
      if(length(unique(mynames1))< length(taxa_names(mbSetObj$dataSet$taxa_table))){
        mynames1 <- make.unique(mynames1, sep="_");
      }
      
      #after cleaning, names might now be NA
      missing.inx <- which(is.na(mynames1))
      missing_vector <- paste0("New_OTU", seq_along(missing.inx))
      mynames1[missing.inx] <- missing_vector
    }
    
    taxa_names(mbSetObj$dataSet$taxa_table)<-taxa_names(data.proc)<-mynames1;
    mbSetObj$dataSet$sample_data<-sample_data(mbSetObj$dataSet$sample_data, errorIfNULL = TRUE);
    mbSetObj$dataSet$proc.phyobj <- merge_phyloseq(data.proc, mbSetObj$dataSet$sample_data, mbSetObj$dataSet$taxa_table);
    
    if(length(rank_names(mbSetObj$dataSet$proc.phyobj)) > 7){
      tax_table(mbSetObj$dataSet$proc.phyobj) <- tax_table(mbSetObj$dataSet$proc.phyobj)[, 1:7]
    }
    
    #also using unique names for our further data
    prefilt.data <- readRDS("data.prefilt");
    rownames(prefilt.data)<-taxa_names(mbSetObj$dataSet$taxa_table);
    saveRDS(prefilt.data, file="data.prefilt")
    
  }else if(mbSetObj$module.type == "sdp"){
    #constructing phyloseq object for aplha diversity.
    data.proc<-otu_table(data.proc, taxa_are_rows = TRUE);
    taxa_names(data.proc)<-rownames(data.proc);
    
    #sanity check: sample names of both abundance and metadata should match.
    smpl_var<-colnames(mbSetObj$dataSet$sample_data);
    indx<-match(colnames(data.proc), rownames(mbSetObj$dataSet$sample_data));
    
    if(all(is.na(indx))){
      current.msg <<-"Please make sure that sample names are matched between both files.";
      return(0);
    }
    mbSetObj$dataSet$sample_data<-sample_data(mbSetObj$dataSet$sample_data, errorIfNULL=TRUE);
    mbSetObj$dataSet$proc.phyobj <- merge_phyloseq(data.proc, mbSetObj$dataSet$sample_data);
  }
  
  saveRDS(mbSetObj$dataSet$proc.phyobj, file="proc.phyobj.orig");
  #do some data cleaning
  mbSetObj$dataSet$data.prefilt <- NULL;
  
  return(.set.mbSetObj(mbSetObj));
}
ApplyAbundanceFilter <- function(mbSetObj, filt.opt, count, smpl.perc){
  
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  data <- readRDS("data.prefilt");
  msg <- NULL;
  
  #this data is used for sample categorial comparision further
  rmn_feat <- nrow(data);
  
  mbSetObj$dataSet$ab.filtered <- FALSE
  
  if(count==0){# no low-count filtering
    kept.inx <- rep(TRUE, rmn_feat);
  }else{
    mbSetObj$dataSet$ab.filtered <- TRUE
    if(filt.opt == "prevalence"){
      rmn_feat <- nrow(data);
      minLen <- smpl.perc*ncol(data);
      kept.inx <- apply(data, MARGIN = 1,function(x) {sum(x >= count) >= minLen});
    }else if (filt.opt == "mean"){
      filter.val <- apply(data, 1, mean, na.rm=T);
      kept.inx <- filter.val >= count;
    }else if (filt.opt == "median"){
      filter.val <- apply(data, 1, median, na.rm=T);
      kept.inx <- filter.val >= count;
    }
  }
  
  mbSetObj$dataSet$filt.data <- data[kept.inx, ];
  saveRDS(mbSetObj$dataSet$filt.data, file="filt.data.orig"); # save an copy
  current.msg <<- paste("A total of ", sum(!kept.inx), " low abundance features were removed based on ", filt.opt, ".", sep="");
  
  return(.set.mbSetObj(mbSetObj));
}

#'Function to filter uploaded data
#'@description This function filters data based on low abundace or variance.
#'Note, this is applied after abundance filter.
#'@param mbSetObj Input the name of the mbSetObj.
#'@param filtopt Character, input the low variance filter option. "iqr" for 
#'inter-quantile range, "sd" for standard deviation, and "cov" for coefficient of variation.
#'@param filtPerct Numeric, input the percentage cutoff for low variance. 
#'@author Jeff Xia \email{jeff.xia@mcgill.ca}
#'McGill University, Canada
#'License: GNU GPL (>= 2)
#'@export

ApplyVarianceFilter <- function(mbSetObj, filtopt, filtPerct){
  
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  data <- mbSetObj$dataSet$filt.data;
  msg <- NULL;
  
  rmn_feat <- nrow(data);
  
  filter.val <- nm <- NULL;
  
  mbSetObj$dataSet$var.filtered <- FALSE
  
  if(filtPerct==0){# no low-count filtering
    remain <- rep(TRUE, rmn_feat);
  }else{
    mbSetObj$dataSet$var.filtered <- TRUE
    if (filtopt == "iqr"){
      filter.val <- apply(data, 1, IQR, na.rm=T);
      nm <- "IQR";
    }else if (filtopt == "sd"){
      filter.val <- apply(data, 1, sd, na.rm=T);
      nm <- "standard deviation";
    }else if (filtopt == "cov"){
      sds <- apply(data, 1, sd, na.rm=T);
      mns <- apply(data, 1, mean, na.rm=T);
      filter.val <- abs(sds/mns);
      nm <- "Coeffecient of variation";
    }
    # get the rank
    rk <- rank(-filter.val, ties.method='random');
    var.num <- nrow(data);
    remain <- rk < var.num*(1-filtPerct);
  }
  
  data <- data[remain,];
  mbSetObj$dataSet$filt.data <- data;
  saveRDS(mbSetObj$dataSet$filt.data, file="filt.data.orig"); # save an copy
  rm.msg1 <- paste("A total of ", sum(!remain), " low variance features were removed based on ", filtopt, ".", sep="");
  rm.msg2 <- paste("The number of features remains after the data filtering step:", nrow(data));
  current.msg <<- paste(c(current.msg, rm.msg1, rm.msg2), collapse=" ");
  mbSetObj$dataSet$filt.msg <- current.msg;
  
  return(.set.mbSetObj(mbSetObj));
  
}
GetMetaTaxaInfo <- function(mbSetObj){
  
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  #check that each rank has >2 groups
  taxa.tbl <- as(tax_table(mbSetObj$dataSet$proc.phyobj), "matrix")
  
  if(ncol(taxa.tbl)==1){
    taxa.nms <- "Phylum"
    return(taxa.nms)
  }
  
  #drop taxa with only 1 level (i.e. Viruses at Phylum)
  gd.inx <- apply(taxa.tbl, 2, function(x) length(unique(x))!=1);
  taxa.tbl.update <- taxa.tbl[,gd.inx];
  taxa.nms <- rank_names(taxa.tbl.update);
  return(taxa.nms[!is.na(taxa.nms)]);
}
UtilMakePhyloseqObjs <- function(mbSetObj, taxrank){
  
  if(mbSetObj$module.type=="mdp"){
    taxa_table <- tax_table(mbSetObj$dataSet$proc.phyobj);
    data <- merge_phyloseq(mbSetObj$dataSet$norm.phyobj, taxa_table);
  }else{
    data <- mbSetObj$dataSet$norm.phyobj; #for shotgun
  }
  
  if(taxrank!="OTU"){
    #merging at taxonomy levels
    data <- fast_tax_glom_mem(data, taxrank);
    if(is.null(data)){
      AddErrMsg("Errors in projecting to the selected taxanomy level!");
      return(0);
    }
  }
  return(data)
}

UtilMakeCountTables <- function(phyloseq.obj, taxrank){
  
  if(taxrank=="OTU"){
    data1 <- as(otu_table(phyloseq.obj), "matrix");
  }else{
    nm <- as.character(tax_table(phyloseq.obj)[,taxrank]);
    #converting NA values to unassigned
    nm[is.na(nm)] <- "Not_Assigned";
    data1 <- as(otu_table(phyloseq.obj), "matrix");
    rownames(data1) <- nm;
    #all NA club together
    data1 <- as.matrix(t(sapply(by(data1, rownames(data1), colSums), identity)));
  }
  return(data1)
}
fast_tax_glom_mem <- function(physeq, taxrank){
  # set up the cache for reuse (i.e updating colors)
  if(!exists("FastTaxGlom_mem")){
    require("memoise");
    FastTaxGlom_mem<<- memoise(fast_tax_glom_first); # the cache will be empty after 5 min
  }
  res <- FastTaxGlom_mem(physeq, taxrank);
  return(res);
}
fast_tax_glom_first <- function(physeq, taxrank){
  
  # setup data. we are going to glob the OTU table based on the Class Taxonomy
  CN  <- which(rank_names(physeq) %in% taxrank);
  if(length(CN) == 0){
    print("name not in the rank!");
    return(NULL);
  }
  tax <- as(access(physeq, "tax_table"), "matrix")[, 1:CN, drop=FALSE];
  tax <- apply(tax, 1, function(i){paste(i, sep=";_;", collapse=";_;")});
  # using Map-Reduce/vectorized
  otab2 <- data.frame(otu_table(physeq))
  taxdf <- data.frame(tax);
  otab2 <- merge(otab2, taxdf, by = "row.names");
  row.names(otab2) <- otab2$Row.names
  otab2 <- otab2[ , 2:ncol(otab2)]
  otab2 <- condenseOTUs(otab2,"tax");
  otab2<-otu_table(otab2,taxa_are_rows = T);
  colnames(otab2)<-sample_names(physeq);
  
  # check if this is meaningful 
  if(sum(is.na(otab2))/length(otab2) > 0.7){
    AddErrMsg("More than 70% values are missing at this level!");
    return(NULL);
  }
  
  if(length(phy_tree(physeq,errorIfNULL = FALSE))==0){
    phy_data<-merge_phyloseq(otab2,tax_table(physeq),sample_data(physeq));
  }else{
    phy_data<-merge_phyloseq(otab2,tax_table(physeq),sample_data(physeq),phy_tree(physeq));
  }
  return(phy_data)
}
condenseOTUs <- function(otutable, splitcol) {
  
  otutable[[splitcol]] <- as.numeric(factor(otutable[[splitcol]], labels=c(1:length(unique(otutable[[splitcol]])))))
  
  # split apply, combine
  splits <- split(otutable, otutable[[splitcol]])
  summed <- Map(colSums, splits)
  summeddf <- Reduce(rbind, summed, init = NULL)
  
  #get rownames.
  # this needs to be changed to get most abundant OTU if thats what is used elsewhere
  # rightnow it returns the first OTU in the group
  newrownames <- Map(function(x){rownames(x)[[1]]}, splits)
  
  #add back rowname and remove tax column
  rownames(summeddf) <- newrownames
  summeddf[, !colnames(summeddf) %in% c(splitcol)]
}
PerformNormalization <- function(mbSetObj, rare.opt, scale.opt, transform.opt){
  mbSetObj <- .get.mbSetObj(mbSetObj);
  data <- readRDS("filt.data.orig");
  tax_nm <- rownames(data);
  msg <- NULL;
  
  if(rare.opt != "none"){
    data <- PerformRarefaction(mbSetObj, data, rare.opt);
    tax_nm <- rownames(data);
    msg <- c(msg, paste("Performed data rarefaction."));
    
    ###### note, rarefying (should?) affect filt.data in addition to norm
    mbSetObj$dataSet$filt.data <- data;
  }else{
    msg <- c(msg, paste("No data rarefaction was performed."));
  }
  
  saveRDS(mbSetObj$dataSet$proc.phyobj, file="orig.phyobj"); # save original phylo.obj
  # now proc.phyobj is now filtered data
  mbSetObj$dataSet$proc.phyobj <- merge_phyloseq(otu_table(data,taxa_are_rows =TRUE), mbSetObj$dataSet$sample_data, mbSetObj$dataSet$taxa_table);
  
  if(scale.opt != "none"){
    if(scale.opt=="colsum"){
      data <- sweep(data, 2, colSums(data), FUN="/")
      data <- data*10000000;
      msg <- c(msg, paste("Performed total sum normalization."));
    }else if(scale.opt=="upperquartile"){
      load_edgeR();
      otuUQ <- edgeRnorm(data,method="upperquartile");
      data <- as.matrix(otuUQ$counts);
      msg <- c(msg, paste("Performed upper quartile normalization"));
    }else if(scale.opt=="CSS"){
      load_metagenomeseq();
      #biom and mothur data also has to be in class(matrix only not in phyloseq:otu_table)
      data1 <- as(data,"matrix");
      dataMR <- newMRexperiment(data1);
      data <- cumNorm(dataMR,p=cumNormStat(dataMR));
      data <- MRcounts(data,norm = T);
      msg <- c(msg, paste("Performed cumulative sum scaling normalization"));
    }else{
      print(paste("Unknown scaling parameter:", scale.opt));
    }
  }else{
    msg <- c(msg, paste("No data scaling was performed."));
  }
  
  if(transform.opt != "none"){
    if(transform.opt=="rle"){
      load_edgeR();            
      otuRLE <- edgeRnorm(data,method="RLE");
      data <- as.matrix(otuRLE$counts);
      msg <- c(msg, paste("Performed RLE Normalization"));
    }else if(transform.opt=="TMM"){
      load_edgeR();   
      otuTMM <- edgeRnorm(data,method="TMM");
      data <- as.matrix(otuTMM$counts);
      msg <- c(msg, paste("Performed TMM Normalization"));
    }else if(transform.opt=="clr"){
      data <- apply(data, 2, clr_transform);
      msg <- "Performed centered-log-ratio normalization.";
    }else{
      print(paste("Unknown scaling parameter:", transform.opt));
    }
  }else{
    msg <- c(msg, paste("No data transformation was performed."));
  }
  #this step has to be done after all the normalization step
  otu.tab <- otu_table(data,taxa_are_rows =TRUE);
  taxa_names(otu.tab) <- tax_nm;
  
  # create phyloseq obj
  mbSetObj$dataSet$sample_data$sample_id <- rownames(mbSetObj$dataSet$sample_data);
  sample_table <- sample_data(mbSetObj$dataSet$sample_data, errorIfNULL=TRUE);
  phy.obj <- merge_phyloseq(otu.tab, sample_table);
  
  #using this object for plotting
  mbSetObj$dataSet$norm.phyobj <- phy.obj;
  current.msg <<- paste(msg, collapse=" ");
  mbSetObj$dataSet$norm.msg <- current.msg;
  
  #make hierarchies
  if(mbSetObj$module.type=="sdp"){
    ranks <- "OTU"
  }else{
    ranks <- c(GetMetaTaxaInfo(mbSetObj), "OTU")
  } 
  
  # start with lowest
  data.list <- list()
  data.list$merged_obj <- vector(length = length(ranks), "list")
  data.list$count_tables <- vector(length = length(ranks), "list")
  names(data.list$count_tables) <- names(data.list$merged_obj) <- ranks
  
  for(i in 1:length(ranks)){
    phyloseq.obj <- UtilMakePhyloseqObjs(mbSetObj, ranks[i])
    data.list$merged_obj[[i]] <- phyloseq.obj
    count.table <- UtilMakeCountTables(phyloseq.obj, ranks[i])
    data.list$count_tables[[i]] <- count.table
  }
  saveRDS(data.list, "phyloseq_objs.RDS")
  
  return(.set.mbSetObj(mbSetObj));
}
load_reshape <- function(){
  suppressMessages(library(reshape))
}
load_viridis <- function(){
  suppressMessages(library(viridis))
}

PlotTaxaAundanceBar<-function(mbSetObj, barplotName, taxalvl, metadata, facet, imgOpt, 
                              feat_cnt, colpalopt, calcmeth, format="png", dpi=72){
  
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  load_reshape();
  load_ggplot();
  load_viridis();
  
  data <- mbSetObj$dataSet$filt.data;
  
  if(class(mbSetObj$dataSet$filt.data)=="matrix"){
    data <- otu_table(data, taxa_are_rows =TRUE);
  }
  
  sample_table <- sample_data(mbSetObj$dataSet$proc.phyobj, errorIfNULL=TRUE);
  data1 <- merge_phyloseq(data, tax_table(mbSetObj$dataSet$proc.phyobj), sample_table);
  data <- as.data.frame(otu_table(data1));
  
  # reorder data based on groups
  sam <- sample_data(data1);
  
  if(facet == "none"){
    sam$newnewnew <- rep("one", nrow(sam));
    facet <- "newnewnew";
  }
  
  sample_data(data1) <- sam
  
  metalp <- as(sample_data(data1), "data.frame") #extract metadata table
  
  smpl_nm <- sample_names(data1);
  clsLbl <- factor(sam[[facet]]);
  
  if(length(levels(clsLbl)) > 9 && min(table(clsLbl)) < 3){
    AddErr("Too many facets to be displayed - please select a more meaningful facet option with at least 3 samples per group.");
    return(0);
  }
  
  ord.inx <- order(clsLbl);
  smpl_nm <- smpl_nm[ord.inx];
  
  clsLbl <- clsLbl[ord.inx];
  colvec <- as.numeric(clsLbl)+1;
  data <- t(data[,ord.inx]);
  
  data_tax <- tax_table(data1);
  
  if(taxalvl=="OTU"){
    taxa_nm <- as.matrix(colnames(data));
    rownames(taxa_nm) <- colnames(data);
    rownames(taxa_nm) <- sub("^X", "", rownames(taxa_nm))
  }else{
    taxa_nm <- data_tax[,taxalvl];
  }
  
  #reshaping data
  taxa_nm <- as.matrix(taxa_nm);
  y <- which(is.na(taxa_nm)==TRUE);
  
  #converting NA values to unassigned; before order it to last position using ZZZ as its name
  taxa_nm[y] <- "ZZZ";
  colnames(data) <- taxa_nm[,1];
  nms <- colnames(data);
  data <- as.matrix(data);
  data <- data %*% sapply(unique(nms),"==",nms);
  data <- data.frame(data);
  data <- data[ , order(names(data))];
  indx <- which(colnames(data)=="ZZZ");
  colnames(data)[indx] <- "NA";
  
  if(calcmeth=="sum"){
    ind <- which(colSums(data)>feat_cnt);
    ind1 <- which(colSums(data)<feat_cnt);
  } else {
    dt <- apply(data,2,median);
    ind <- which(dt>feat_cnt);
    ind1 <- which(dt<feat_cnt);
  }
  
  if(length(ind)==0){
    current.msg<<-"All features have lower read count than given minimum count filter. Please lower the cut off for minimum count.";
    return(0);
  }
  
  if(length(ind1)>0){
    colnames(data)[ind1] <- "Others";
    data <- as.data.frame(do.call(cbind, by(t(data),INDICES=names(data),FUN=colSums)));
  }
  
  yLbl <- "Actual Abundance";
  
  if(imgOpt=="barnorm"){
    data <- as.data.frame(apply(data,1, function(x) x/sum(x)));
    data<-as.data.frame(t(data));
    yLbl <- "Relative Abundance";
  }
  
  # height according to number of legends
  feat_no <- ncol(data);
  h <- 550;
  if(feat_no < 10){
    h <- h;
  } else if (feat_no < 20){
    h <- h+100;
  } else if (feat_no < 50){
    h <- h+200;
  } else if (feat_no < 100){
    h <- h+400;
  } else if (feat_no > 100){
    h <- h+500;
  }
  
  # width calculation
  a <- nsamples(data1);
  min.w <- 540;
  if(length(a)<50){
    w <- a*35;
  }else{
    w <- a*25;
  }
  if(w < min.w){
    w <- min.w;
  }
  
  data <- data[row.names(metalp), ]
  data[[get("facet")]] <- metalp[[get("facet")]]
  data$sample <- row.names(data);
  write.csv(t(data), file="taxa_abund.csv");
  data <- melt(data, id = c("sample", get("facet")))
  tmp_df <- aggregate(data$value, by=list(data$variable), FUN=mean)
  var_level <- tmp_df[order(tmp_df$x, decreasing = TRUE), ][[1]]
  as.character()# get the order of taxa relative abundance
  
  data$variable <- factor(data$variable, levels = var_level) # change the factor level of taxa
  levels(data$variable) <- sub("^X", "", levels(data$variable)) 
  
  #color schema
  color_var <- levels(factor(data$variable));
  x <- length(color_var);
  
  if(colpalopt=="grad"){
    indx <- which(color_var=="NA");
    #color schema for ggplot
    x.colors <- hcl(h=seq(15,375,length=(x+1)),l=65,c=100)[1:x];
    x.colors[indx] <- "#666666";
  }else if (colpalopt=="cont21"){
    x.colors <- rep(custom_col21,length.out=x);
  }else if (colpalopt=="cont28"){
    x.colors <- rep(custom_final28,length.out=x);
  }else {
    x.colors <- rep(custom_col42,length.out=x);
  }
  
  barplotName = paste(barplotName, ".",format, sep="");
  mbSetObj$imgSet$stack <- barplotName;
  
  Cairo::Cairo(file=barplotName,width=w, height=h, type=format, bg="white",dpi=dpi);
  
  if(length(unique(data$sample)) <= 10){
    guide_num = 3
  }else if (length(unique(data$sample)) <= 20){
    guide_num = 4
  }else{
    guide_num = 5
  }
  
  box <- ggplot(data = data,
                aes(x = sample,
                    y = value,
                    fill = variable)) +
    geom_bar(stat="identity", position="stack") +
    facet_grid(as.formula(paste("~", facet)), scales = "free", space = "free") +
    labs(y=yLbl,x="", fill = taxalvl)+
    theme(legend.position="bottom",legend.box = "vertical",
          axis.text.x = element_text(angle = 45,
                                     hjust =1,vjust=1), #remove colour=colvec
          axis.title.x=element_blank(),
          panel.background = element_rect(fill = "grey90",
                                          colour = "grey90",
                                          size = 0.5, linetype = "solid"),
          axis.line = element_line(colour = "lightgrey"));
  
  if(colpalopt=="set3"){
    cols.needed <- length(unique(data$variable))
    if(cols.needed > 12){
      col.func <- grDevices::colorRampPalette(RColorBrewer::brewer.pal(12, "Set3"))
      box <- box + scale_fill_manual(values=col.func(cols.needed), guide = guide_legend(direction = "horizontal", ncol=guide_num))
    }else{
      box <- box + scale_fill_brewer(palette = "Set3", guide = guide_legend(direction = "horizontal", ncol=guide_num))
    }
  }else if(colpalopt=="viridis"){
    box <- box + viridis::scale_fill_viridis(discrete=TRUE) + guides(color=guide_legend(ncol=guide_num)) 
  }else if(colpalopt %in% c("magma","plasma","inferno")){
    box <- box + viridis::scale_fill_viridis(option=colpalopt, discrete=TRUE) + guides(color=guide_legend(ncol=guide_num))
  }else{
    box <- box + scale_fill_manual(values=c(x.colors)) + guides(color=guide_legend(ncol=guide_num))
  }
  
  if(facet == "newnewnew"){
    box <- box + theme(strip.text.x = element_blank())
  }
  
  print(box);
  dev.off();
  mbSetObj$analSet$stack <- data;
  mbSetObj$analSet$stack.taxalvl <- taxalvl;
  mbSetObj$analSet$plot <- "Stacked Bar";
  
  return(.set.mbSetObj(mbSetObj));
}
PlotAlphaData<-function(mbSetObj, data.src, bargraphName, distName, metadata, taxrank, colors = "default", format="png", dpi=72){
  
  mbSetObj <- .get.mbSetObj(mbSetObj);  
  set.seed(13133);
  
  if(data.src == "orig"){
    data <- readRDS("orig.phyobj");
  }else{
    data <- mbSetObj$dataSet$proc.phyobj;
  }
  
  if(taxrank!="OTU"){
    #merging at taxonomy levels
    data <- fast_tax_glom_mem(data, taxrank);
    if(is.null(data)){
      AddErrMsg("Errors in projecting to the selected taxanomy level!");
      return(0);
    }
  } 
  
  #reordering the sample (getting reordered rownames)
  sam <- sample_data(data);
  sam <- sam[order(sam[[metadata]])];
  smplord <- rownames(sam);
  
  smpl.num <- length(smplord);
  if(smpl.num < 25){
    width <- 600
  }else if(smpl.num >= 25 & smpl.num <= 50){
    width <- 750
  }else{
    width <- 900
  }
  
  bargraphName = paste(bargraphName, ".", format, sep="");
  mbSetObj$imgSet$alpha <- bargraphName;  
  Cairo::Cairo(file=bargraphName, width, height=450, type=format, bg="white", dpi=dpi);
  box = plot_richness(data, color = metadata, measures = distName) + scale_x_discrete(limits=c(smplord));
  
  if(colors == "viridis"){
    box = box + viridis::scale_color_viridis(discrete=TRUE)
  }else if(colors %in% c("magma","plasma","inferno")){
    box <- box + viridis::scale_color_viridis(option=colors, discrete=TRUE)
  }
  
  mbSetObj$analSet$alpha <- box$data;
  write.csv(mbSetObj$analSet$alpha, file="alphadiversity.csv");
  box = box + theme_bw() + theme(axis.text.x = element_text(angle = 45, hjust = 1,vjust= 1));
  box$layers <- box$layers[-1];
  box <- box + geom_point(size=3, alpha=0.7);
  #getting scale for plot (using same for boxplot also)
  ylimits <<- layer_scales(box)$y$range$range;
  print(box);
  
  dev.off();
  mbSetObj$analSet$alpha.meth <- distName;
  mbSetObj$analSet$alpha.metadata <- metadata;
  mbSetObj$analSet$alpha.taxalvl <- taxrank;
  return(.set.mbSetObj(mbSetObj))
}
PlotAlphaBoxData<-function(mbSetObj, boxplotName, distName, metadata, colors="default", format="png", dpi=72){
  
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  load_viridis();
  
  set.seed(1313397);
  data <- mbSetObj$analSet$alpha;
  CLASS <- data[,metadata];
  boxplotName = paste(boxplotName, ".", format, sep="");
  mbSetObj$imgSet$alpha.box<-boxplotName;
  
  grp_size <- length(levels(CLASS))
  
  if(grp_size <= 2){
    width <- 300
  }else if(grp_size >= 3 & grp_size < 6){
    width <- 400
  }else{
    width <- 500
  }
  
  Cairo::Cairo(file=boxplotName, width, height=400, type=format, bg="white", dpi=dpi);
  
  if(colors %in% c("magma","plasma","inferno","viridis")){
    box1 = ggplot(data, aes(data[,metadata], data$value, fill = CLASS))
  }else{
    box1 = ggplot(data, aes(data[,metadata], data$value, color = CLASS))
  }
  
  box1 = box1 + stat_boxplot(geom ='errorbar', width=0.2) +
    geom_boxplot(alpha=0.7, aes(outlier.shape=1),
                 position = position_dodge(width = 0), width=0.3) +
    geom_jitter(width = 0.1) + #reduce the width of jitter
    stat_summary(fun.y=mean, #add mean point
                 geom = "point",
                 shape = 18,
                 size = 4,
                 color = "black") +
    theme_bw() +
    theme(axis.text.x = element_text(angle = 45, vjust = 1, hjust = 1, size = 11),
          axis.text.y = element_text(size = 11),
          legend.text = element_text(size = 11)) + #adjust titles
    labs(title = "",
         y= paste("Alpha-diversity Index:", as.character(data$variable[[1]]), sep = " "),
         x="") + #remove x = CLASS, add title name, change y name
    theme(axis.title.y = element_text(size=14)) + 
    coord_cartesian(ylim = c(ylimits[1], ylimits[2]));
  
  if(colors == "viridis"){
    box1 = box1 + viridis::scale_fill_viridis(discrete = TRUE);
  }else if(colors %in% c("magma","plasma","inferno")){
    box1 <- box1 + viridis::scale_fill_viridis(option=colors, discrete=TRUE)
  }
  
  print(box1);
  dev.off();
  return(.set.mbSetObj(mbSetObj))
}
load_datatable <- function(){
  suppressMessages(library(data.table))
}
PlotBetaDiversity<-function(mbSetObj, plotNm, ordmeth, distName, colopt, metadata, 
                            showlabel, taxrank, taxa, alphaopt, ellopt, format="png", dpi=72,
                            custom_col = "none"){
  
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  load_datatable();
  load_viridis();
  
  set.seed(13134);
  metadata <- metadata;
  
  #using normalized data
  if(taxrank=="OTU"){
    taxa_table <- tax_table(mbSetObj$dataSet$proc.phyobj);
    data <- merge_phyloseq(mbSetObj$dataSet$norm.phyobj, taxa_table);
  }else{
    taxa_table <- tax_table(mbSetObj$dataSet$proc.phyobj);
    data <- merge_phyloseq(mbSetObj$dataSet$norm.phyobj, taxa_table);
    #merging at taxonomy levels
    data <- fast_tax_glom_mem(data,taxrank);
    if(is.null(data)){
      AddErrMsg("Errors in projecting to the selected taxanomy level!");
      return(0);
    }
  }
  
  if(colopt=="taxa"){
    if(taxrank=="OTU"){
      data1 <- as.matrix(otu_table(data));
      feat_data <- as.numeric(data1[taxa,]);
    }else{
      nm <- as.character(tax_table(data)[,taxrank]);
      if(sum(is.na(nm))/length(nm) > 0.7){
        AddErrMsg("More than 70% values are missing at this taxa!");
        return(0);
      }
      #converting NA values to unassigned
      nm[is.na(nm)] <- "Not_Assigned";
      data1 <- as.matrix(otu_table(data));
      rownames(data1) <- nm;
      #all NA club together
      data1 <- as.matrix(t(sapply(by(data1, rownames(data1), colSums), identity)));
      feat_data <- data1[taxa,];
    }
    sample_data(data)$taxa <- feat_data;
    indx <- which(colnames(sample_data(data))=="taxa");
    colnames(sample_data(data))[indx] <- taxa;
    taxa1 <- colnames(sample_data(data))[indx];
    taxaorig <- taxa;
    #if the taxa names are numeric then X is appending to column name
    
    if(!is.na(as.numeric(taxa))=="TRUE"){
      taxa <- paste("X",taxa, sep = "", collapse = NULL);
    }
  }else if(colopt=="alphadiv") {
    data1 <- mbSetObj$dataSet$proc.phyobj;
    box <- plot_richness(data1, measures = alphaopt);
    alphaboxdata <- box$data;
    sam_nm <- sample_names(data);
    alphaboxdata <- alphaboxdata[alphaboxdata$samples %in% sam_nm,];
    alphaval <- alphaboxdata$value;
    sample_data(data)$alphaopt <- alphaval;
    indx <- which(colnames(sample_data(data))=="alphaopt");
    colnames(sample_data(data))[indx] <- alphaopt;
  }else{
    data<-data;
  }
  
  if(distName=="wunifrac"){
    
    load_ape();
    pg_tree <- readRDS("tree.RDS");
    pg_tb <- tax_table(data);
    pg_ot <- otu_table(data);
    pg_sd <- sample_data(data);
    pg_tree <- prune_taxa(taxa_names(pg_ot), pg_tree);
    data <- merge_phyloseq(pg_tb, pg_ot, pg_sd, pg_tree);
    
    saveRDS(data, "data_unifra.RDS");
    ord <- ordinate(data,method = ordmeth,"unifrac",weighted=TRUE);
    
  } else if (distName=="unifrac") {
    
    load_ape();      
    pg_tree <- readRDS("tree.RDS");
    pg_tb <- tax_table(data);
    pg_ot <- otu_table(data);
    pg_sd <- sample_data(data);
    pg_tree <- prune_taxa(taxa_names(pg_ot), pg_tree);
    data <- merge_phyloseq(pg_tb, pg_ot, pg_sd, pg_tree);
    
    saveRDS(data, "data_unifra.RDS");
    ord<-ordinate(data,method = ordmeth,"unifrac");
    
  }else{
    ord<-ordinate(data,method = ordmeth,distName);
  }
  
  plotNm = paste(plotNm, ".", format, sep="");
  mbSetObj$imgSet$beta2d<-plotNm;
  
  Cairo::Cairo(file=plotNm, width=720, height=500, type=format, bg="white",dpi=dpi);
  
  if(colopt=="taxa"){
    box = plot_ordination(data, ord, color=taxa) + labs(aesthetic=taxaorig) + scale_colour_gradient(low="green", high="red");
  }else if(colopt=="alphadiv") {
    box = plot_ordination(data, ord, color=alphaopt)+ scale_colour_gradient(low="green", high="red");
  }else{
    box = plot_ordination(data, ord, color=metadata);
  }
  
  box$layers <- box$layers[-1];
  
  if(showlabel=="samnm"){
    box = box + geom_text(aes(label=sample_id), hjust=0.5, vjust=2, size=3, fontface="bold");
    box = box + geom_point(size=4, alpha=0.6) + theme_bw();
  }else if(showlabel=="none"){
    box=box+geom_point(size=4, alpha=0.8) + theme_bw();
  }else{
    showlabel <<- showlabel;
    bx_data <<- data.frame(box$data);
    box = box + geom_text(aes(label=bx_data[ ,showlabel]), hjust=0.5, vjust=2, size=3, fontface="bold");
    box = box + geom_point(size=4, alpha=0.6) + theme_bw();
  }
  
  #used for area color for ellipse
  if(colopt=="expfac"){
    sam_data <- as.data.frame(sample_data(data));
    clsLbl <- sam_data[[metadata]];
    
    if (ellopt=="yes"){
      box = box + stat_ellipse(type="norm", linetype=2, geom = "polygon", alpha = 0.2, aes_string(fill = clsLbl), show.legend=FALSE);
    }
  }
  
  if(custom_col == "viridis"){
    box = box + viridis::scale_color_viridis(discrete = TRUE) + viridis::scale_fill_viridis(discrete = TRUE)
  }else if(custom_col == "plasma"){
    box = box + viridis::scale_color_viridis(option="plasma", discrete = TRUE) + viridis::scale_fill_viridis(option="plasma", discrete = TRUE)
  }else if(custom_col == "magma"){
    box = box + viridis::scale_color_viridis(option="magma", discrete = TRUE) + viridis::scale_fill_viridis(option="magma", discrete = TRUE)
  }
  
  print(box);
  dev.off();
  
  #saving info for report generation
  mbSetObj$analSet$beta <- data;
  mbSetObj$analSet$beta.meth <- ordmeth;
  mbSetObj$analSet$beta.dist <- distName;
  mbSetObj$analSet$beta.taxalvl <- taxrank;
  
  # if it is NMDS, show stress value in the plot
  if(ordmeth == "NMDS"){
    mbSetObj$analSet$beta.stress <- paste("[NMDS] Stress =", signif(ord$stress, 5));
  }
  
  return(.set.mbSetObj(mbSetObj))
}
PlotRarefactionCurve <- function(mbSetObj, data.src, linecolor, linetype, facet, step, 
                                 imgName, format="png",dpi=72){
  mbSetObj <- .get.mbSetObj(mbSetObj);
  # should use unfiltered data
  if(data.src == "orig"){
    data_rare <- readRDS("orig.phyobj");
  }else{
    data_rare <- mbSetObj$dataSet$proc.phyobj;
  }
  
  sample_table_msg <- sample_data(data_rare, errorIfNULL=TRUE);
  if(min(table(factor(sample_table_msg [[linecolor]]))) < 3 | min(table(factor(sample_table_msg [[linetype]]))) < 3 | min(table(factor(sample_table_msg [[facet]]))) < 3){
    current.msg<<-"Too many groups to be displayed - please select a more meaningful group option with at least 3 samples per group.";
    return(0);
  }
  
  #get good's coverage index
  #goods_coverage <- ComputeGoods(data_rare)
  write.csv(goods_coverage, "goods_coverage.csv", row.names = FALSE, quote = FALSE);
  
  feat_no <- nsamples(data_rare);
  
  #adjust height according to number of legends
  w <- 600;
  
  if(feat_no < 10){
    w<-w;
  } else if (feat_no < 20){
    w<-w+100;
  } else if (feat_no < 50){
    w<-w+200;
  } else if (feat_no < 100){
    w<-w+400;
  } else if (feat_no > 100){
    w<-w+500;
  }
  
  imgName = paste(imgName,".", format, sep="");
  Cairo::Cairo(file=imgName, width = 1.5 * w, height = 540, type=format, bg="white", dpi=dpi);
  linecolor <- ifelse(linecolor == "none", "NULL", linecolor);
  linetype <- ifelse(linetype == "none", "NULL", linetype);
  
  box <- ggrare2(data_rare,
                 data.src = data.src,
                 color = linecolor,
                 label = "Sample",
                 linetype = linetype,
                 se = FALSE,  # this is not to meaningful
                 step = step);
  
  if(!is.null(facet) & facet != "none"){
    box <- box + facet_wrap(as.formula(paste("~", facet)))
  } else {
    box <- box;
  }
  print(box);
  dev.off();
  mbSetObj$analSet$rarefaction_curve <- imgName;
  mbSetObj$analSet$rarefaction_curve_data.src = ifelse(data.src == "orig", "original", "filtered");
  return(.set.mbSetObj(mbSetObj))
}
ggrare2 <- function(physeq_object, data.src, label = NULL, color = NULL, plot = TRUE, linetype = NULL, se = FALSE, step=5) {
  
  x <- methods::as(otu_table(physeq_object), "matrix")
  
  if (taxa_are_rows(physeq_object)) { x <- t(x) }
  
  ## This script is adapted from vegan `rarecurve` function
  tot <- rowSums(x)
  S <- rowSums(x > 0)
  nr <- nrow(x)
  step_new = floor(max(tot)/as.integer(step))
  rarefun <- function(i) {
    #cat(paste("rarefying sample", rownames(x)[i]), sep = "\n")
    n <- seq(1, tot[i], by = step_new)
    if (n[length(n)] != tot[i]) {
      n <- c(n, tot[i])
    }
    y <- vegan::rarefy(x[i, ,drop = FALSE], n, se = se)
    if (nrow(y) != 1) {
      rownames(y) <- c(".S", ".se")
      return(data.frame(t(y), Size = n, Sample = rownames(x)[i]))
    } else {
      return(data.frame(.S = y[1, ], Size = n, Sample = rownames(x)[i]))
    }
  }
  
  f_n <- paste(data.src, step, "rds", sep = ".");
  
  if(file.exists(f_n)){
    df <- readRDS(file = f_n);
  } else {
    out <- lapply(seq_len(nr), rarefun)
    df <- do.call(rbind, out);
    saveRDS(df, file = f_n);
  }
  
  # Get sample data
  if (!is.null(sample_data(physeq_object, FALSE))) {
    sdf <- methods::as(sample_data(physeq_object), "data.frame")
    sdf$Sample <- rownames(sdf)
    data <- merge(df, sdf, by = "Sample")
    labels <- data.frame(x = tot, y = S, Sample = rownames(x))
    labels <- merge(labels, sdf, by = "Sample")
  }
  
  
  # Add, any custom-supplied plot-mapped variables
  if (length(color) > 1) {
    data$color <- color
    names(data)[names(data) == "color"] <- deparse(substitute(color))
    color <- deparse(substitute(color))
  }
  
  if (length(label) > 1) {
    labels$label <- label
    names(labels)[names(labels) == "label"] <- deparse(substitute(label))
    label <- deparse(substitute(label))
  }
  
  if (length(linetype) > 1) {
    data$linetype <- linetype
    names(data)[names(data) == "linetype"] <- deparse(substitute(linetype))
    linetype <- deparse(substitute(linetype))
  }
  
  p <- ggplot2::ggplot(data = data,
                       ggplot2::aes_string(x = "Size",
                                           y = ".S",
                                           group = "Sample",
                                           color = color,
                                           linetype = linetype))
  
  p <- p + ggplot2::labs(x = "Sequence Sample Size", y = "Species Richness")
  
  if (!is.null(label)) {
    p <- p + ggplot2::geom_text(data = labels,
                                ggplot2::aes_string(x = "x",
                                                    y = "y",
                                                    label = label,
                                                    color = color),
                                size = 4, hjust = 0) +
      scale_x_continuous(expand = c(0, 0, 0.3, 0))
  }
  
  p <- p + ggplot2::geom_line()
  if (se) { ## add standard error if available
    p <- p +
      ggplot2::geom_ribbon(ggplot2::aes_string(ymin = ".S - .se",
                                               ymax = ".S + .se",
                                               color = NULL,
                                               fill = color),
                           alpha = 0.2)
  }
  invisible(p)
}
load_pheatmap <- function(){
  suppressMessages(library(pheatmap))
}
load_rcolorbrewer <- function(){
  suppressMessages(library(RColorBrewer))
}
GetColorSchema <- function(mbSetObj, grayscale=F){
  
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  # test if total group number is over 9
  claslbl <- as.factor(sample_data(mbSetObj$dataSet$norm.phyobj)[[variable]]);
  grp.num <- length(levels(claslbl));
  
  if(grayscale){
    dist.cols <- grDevices::colorRampPalette(c("grey90", "grey30"))(grp.num);
    lvs <- levels(claslbl);
    colors <- vector(mode="character", length=length(claslbl));
    
    for(i in 1:length(lvs)){
      colors[mbSetObj$analSet$cls == lvs[i]] <- dist.cols[i];
    }
  }else if(grp.num > 9){
    pal12 = c("#A6CEE3", "#1F78B4", "#B2DF8A", "#33A02C", "#FB9A99",
              "#E31A1C", "#FDBF6F", "#FF7F00", "#CAB2D6", "#6A3D9A",
              "#FFFF99", "#B15928");
    dist.cols <- grDevices::colorRampPalette(pal12)(grp.num);
    lvs <- levels(claslbl);
    colors <- vector(mode="character", length=length(mbSetObj$analSet$cls));
    
    for(i in 1:length(lvs)){
      colors[claslbl == lvs[i]] <- dist.cols[i];
    }
  }else{
    if(exists("colVec") && !any(colVec =="#NA") ){
      cols <- vector(mode="character", length=length(claslbl));
      clsVec <- as.character(claslbl);
      grpnms <- names(colVec);
      
      for(i in 1:length(grpnms)){
        cols[clsVec == grpnms[i]] <- colVec[i];
      }
      colors <- cols;
    }else{
      colors <- as.numeric(claslbl)+1;
    }
  }
  return (colors);
}

PlotHeatmap<-function(mbSetObj, plotNm, smplDist, clstDist, palette, metadata,
                      taxrank, viewOpt, doclust, format="png", showfeatname,
                      appendnm, rowV=F, colV=T, var.inx=NA, border=T, width=NA, dpi=72){
  
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  load_pheatmap();
  load_rcolorbrewer();
  load_viridis();
  
  set.seed(2805614);
  #used for color pallete
  variable <<- metadata;
  data <- mbSetObj$dataSet$norm.phyobj;
  
  if(mbSetObj$module.type=="mdp"){
    mbSetObj$dataSet$taxa_table <- tax_table(mbSetObj$dataSet$proc.phyobj);
    data <- merge_phyloseq(data, mbSetObj$dataSet$taxa_table);
  }else{
    taxrank <- "OTU";
  }
  
  #if more than 1500 features will be present;subset to most abundant=>1500 features.
  #OTUs already in unique names;
  if(ntaxa(data)>1500){
    data = prune_taxa(names(sort(taxa_sums(data), TRUE))[1:1500], data);
    viewOpt == "overview";
  }
  
  if(taxrank=="OTU"){
    data1 <- as.matrix(otu_table(data));
    rownames(data1) <- taxa_names(data);
  }else{
    #merging at taxonomy levels
    data <- fast_tax_glom_mem(data,taxrank);
    if(is.null(data)){
      AddErrMsg("Errors in projecting to the selected taxanomy level!");
      return(0);
    }
    nm <- as.character(tax_table(data)[,taxrank]);
    y <- which(is.na(nm)==TRUE);
    #converting NA values to unassigned
    nm[y] <- "Not_Assigned";
    data1 <- as.matrix(otu_table(data));
    
    if(appendnm=="T"){
      all_nm <- colnames(tax_table(data));
      hg_nmindx <- which(all_nm==taxrank)-1;
      
      if(hg_nmindx!=0){
        nma <- as.character(tax_table(data)[,hg_nmindx]);
        y1 <- which(is.na(nma)==TRUE);
        nma[y1] <- "Not_Assigned";
        nm <- paste0(nma,"_",nm);
        ind <- which(nm=="Not_Assigned_Not_Assigned");
        nm[ind] <- "Not_Assigned";
        nm <- gsub("_Not_Assigned", "",nm, perl = TRUE);
      }
    }
    
    rownames(data1) <- nm;
    #all NA club together
    data1 <- (t(sapply(by(data1,rownames(data1),colSums),identity)));
    nm <- rownames(data1);
  }
  
  # arrange samples on the basis of slected experimental factor and using the same for annotation also
  annotation <- data.frame(sample_data(data));
  
  ind <- which(colnames(annotation)!=metadata && colnames(annotation)!="sample_id");
  
  if(length(ind)>0){
    ind1 <- ind[1];
    annotation <- annotation[order(annotation[,metadata],annotation[,ind1]),];
  }else{
    annotation <- annotation[order(annotation[,metadata]),];
  }
  
  # remove those columns that all values are unique (continuous or non-factors)
  #uniq.inx <- apply(annotation, 2, function(x){length(unique(x)) == length(x)});
  #there is an additional column sample_id which need to be removed first
  # get only good meta-data
  good.inx <- GetDiscreteInx(annotation);
  if(sum(good.inx)>0){
    annotation <- annotation[,good.inx, drop=FALSE];
    sam.ord <- rownames(annotation);
    data1 <- data1[,sam.ord];
  }else{
    annotation <- NA;
  }
  
  # set up colors for heatmap
  if(palette=="gbr"){
    colors <- grDevices::colorRampPalette(c("green", "black", "red"), space="rgb")(256);
  }else if(palette == "heat"){
    colors <- grDevices::heat.colors(256);
  }else if(palette == "topo"){
    colors <- grDevices::topo.colors(256);
  }else if(palette == "gray"){
    colors <- grDevices::colorRampPalette(c("grey90", "grey10"), space="rgb")(256);
  }else if(palette == "byr"){
    colors <- rev(grDevices::colorRampPalette(RColorBrewer::brewer.pal(10, "RdYlBu"))(256));
  }else if(palette == "viridis") {
    colors <- rev(viridis::viridis(10))
  }else if(palette == "plasma") {
    colors <- rev(viridis::plasma(10))
  }else {
    colors <- rev(grDevices::colorRampPalette(RColorBrewer::brewer.pal(10, "RdBu"))(256));
  }
  
  if(showfeatname=="T"){
    showfeatname<-T;
    min.margin <- 360;
  } else {
    showfeatname<-F;
    min.margin <- 200;
  }
  
  #setting the size of plot
  if(is.na(width)){
    minW <- 800;
    myW <- ncol(data1)*20 + min.margin;
    if(myW < minW){
      myW <- minW;
    }
    w <- round(myW/72,2);
  }
  
  myH <- nrow(data1)*20 + 180;
  h <- round(myH/72,2);
  
  if(viewOpt == "overview"){
    if(is.na(width)){
      if(w >9.3){
        w <- 9.3;
      }
    }
    if(h > w){
      h <- w;
    }
  }
  
  if(border){
    border.col<-"grey60";
  }else{
    border.col <- NA;
  }
  
  plotNm = paste(plotNm, ".", format, sep="");
  mbSetObj$imgSet$heatmap<-plotNm;
  
  if(format=="pdf"){
    grDevices::pdf(file = plotNm, width=w, height=h, bg="white", onefile=FALSE);
  }else{
    Cairo::Cairo(file = plotNm, unit="in", dpi=dpi, width=w, height=h, type=format, bg="white");
  }
  
  # set up color schema for samples
  if(palette== "gray"){
    cols <- GetColorSchema(mbSetObj, T);
    uniq.cols <- unique(cols);
  }else{
    cols <- GetColorSchema(mbSetObj, F);
    uniq.cols <- unique(cols);
  }
  
  if(doclust=="T"){
    rowV<-T;
  }
  
  pheatmap::pheatmap(data1,
                     annotation=annotation,
                     fontsize=8, fontsize_row=8,
                     clustering_distance_rows = smplDist,
                     clustering_distance_cols = smplDist,
                     clustering_method = clstDist,
                     show_rownames = showfeatname,
                     border_color = border.col,
                     cluster_rows = colV,
                     cluster_cols = rowV,
                     scale= "row",
                     color = colors
  );
  
  dev.off();
  
  # storing for Report Generation
  mbSetObj$analSet$heatmap<-data1;
  mbSetObj$analSet$heatmap.dist<-smplDist;
  mbSetObj$analSet$heatmap.clust<-clstDist;
  mbSetObj$analSet$heat.taxalvl<-taxrank;
  return(.set.mbSetObj(mbSetObj))
}

PerformMetagenomeSeqAnal<-function(mbSetObj, variable, p.lvl, shotgunid, taxrank, model){
  
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  load_metagenomeseq();
  
  filt.dataphy <- mbSetObj$dataSet$filt.data;
  filt.dataphy <- apply(filt.dataphy,2,as.integer);
  filt.dataphy <- otu_table(filt.dataphy,taxa_are_rows =TRUE);
  sample_table <- sample_data(mbSetObj$dataSet$norm.phyobj, errorIfNULL=TRUE);
  filt.dataphy <- merge_phyloseq(filt.dataphy, sample_table);
  taxa_names(filt.dataphy) <- rownames(mbSetObj$dataSet$filt.data);
  data <- filt.dataphy;
  
  #data<-dataSet$norm.phyobj;
  cls <- as.factor(sample_data(data)[[variable]]);
  lvl <- length(levels(cls));
  
  if(mbSetObj$module.type=="mdp"){
    mbSetObj$dataSet$taxa_table <- tax_table(mbSetObj$dataSet$proc.phyobj);
    data <- merge_phyloseq(data, mbSetObj$dataSet$taxa_table);
  }else{ #using by default names for shotgun data
    taxrank <- "OTU";
  }
  
  if(taxrank!="OTU"){
    #merging at taxonomy levels
    data <- fast_tax_glom_mem(data, taxrank);
    if(is.null(data)){
      AddErrMsg("Errors in projecting to the selected taxanomy level!");
      return(0);
    }
    nm <- as.character(tax_table(data)[,taxrank]);
    #converting NA values to unassigned
    nm[is.na(nm)] <- "Not_Assigned";
    data1 <- as.matrix(otu_table(data));
    rownames(data1) <- nm;
    #all NA club together
    data1 <- as.matrix(t(sapply(by(data1,rownames(data1),colSums),identity)));
    data1 <- otu_table(data1, taxa_are_rows=T);
    data <- merge_phyloseq(data1, sample_data(data));
    nm <- taxa_names(data);
  }
  
  tree_data <<- data;
  data <- phyloseq_to_metagenomeSeq(data);
  data <- cumNorm(data, p=cumNormStat(data));
  mod <- model.matrix(~phenoData(data)@data[,variable]);
  
  if(model=="zigfit"){
    fit <- fitZig(data, mod);  
  }else{
    if(length(levels(cls)) > 2){
      current.msg <<- paste( "More than two groups present in your experimental factor. This model can only be used with two groups.");
      return(0);
    }else{
      fit <-fitFeatureModel(data, mod);
    }
  }
  
  x <- MRfulltable(fit, number = nrow(assayData(data)$counts));
  x <- x[!is.na(rownames(x)), ];
  
  rownames(x) <- gsub(":1", "", x = rownames(x), fixed = TRUE);
  x$OTUnames <- as.character(rownames(x))
  
  if (!is.null(tax_table(data, errorIfNULL = FALSE))) {
    #Attach the bacterial taxonomy to the table, if available
    TAX = data.frame(tax_table(data));
    TAX$OTUnames <- as.character(rownames(TAX));
    res = merge(x, TAX, by = "OTUnames")
  } else {
    res = x;
  }
  
  # Sort and return #sighits return TRUE or FALSE
  sigHits <-res$adjPvalues<=p.lvl;
  de.Num <- length(which(sigHits));
  
  if(de.Num == 0){
    current.msg <<- paste( "No significant features were identified using the given p value cutoff. Please change the cutoff limit.");
  }else{
    current.msg <<- paste("A total of", de.Num, "significant features were identified!")
  }
  
  if(model=="ffm"){
    resTable <- res[,c("pvalues","adjPvalues","logFC")];
    resTable <- signif(resTable[,c(3,1,2)], digits = 5);
    colnames(resTable) <- c("logFC","Pvalues","FDR");
  }else{
    resTable <-res[,c("pvalues","adjPvalues")];
    resTable <- signif(resTable[,1:2], digits = 5);
    colnames(resTable) <- c("Pvalues","FDR");
  }
  
  ord.inx <- order(resTable$Pvalues);
  resTable <- resTable[ord.inx, , drop=FALSE];
  write.csv(resTable, file="metageno_de_output.csv");
  
  if(nrow(resTable) > 500){
    resTable <- resTable[1:500, ];
  }
  
  mbSetObj$analSet$metagenoseq$resTable <- mbSetObj$analSet$resTable <- data.frame(resTable);
  
  #only getting the names of DE features
  diff_ft <<- rownames(resTable)[1:de.Num];
  sigfeat <- rownames(resTable);
  
  #prepare individual boxplot
  box_data <- MRcounts(data);
  #subset only diff. Abundant features
  box_data <- box_data[sigfeat, ];
  
  #samples in rows
  box_data <- t(box_data);
  box_data <- data.frame(box_data);
  colnames(box_data) <- sigfeat;
  
  claslbl <- pData(data)[ ,variable];
  box_data$class <- unlist(claslbl);
  
  mbSetObj$analSet$boxdata <- box_data;
  mbSetObj$analSet$sig.count <- de.Num;
  mbSetObj$analSet$anal.type <- "metagseq";
  mbSetObj$analSet$var.type <- variable;
  mbSetObj$analSet$metageno.taxalvl <- taxrank;
  mbSetObj$analSet$id.type <- shotgunid;
  
  return(.set.mbSetObj(mbSetObj));
}
load_metagenomeseq <- function(){
  suppressMessages(library(metagenomeSeq))
}

RF.Anal <- function(mbSetObj, treeNum, tryNum, randomOn, variable, taxrank){
  
  load_randomforest();
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  # set up random numbers
  if(is.null(mbSetObj$analSet$random.seeds)){
    mbSetObj$analSet$random.seeds <- GetRandomNumbers();
    mbSetObj$analSet$cur.inx <- 0;
    mbSetObj$analSet$rn.seed <- mbSetObj$analSet$random.seeds[1];
  }
  
  if(randomOn == -1){
    rn.sd <- 123456;
  }else if(randomOn == 0){ # keep current
    rn.sd <- mbSetObj$analSet$rn.seed;
  }else{ # random on
    cur.inx <- mbSetObj$analSet$cur.inx + 1;
    rn.sd <- mbSetObj$analSet$random.seeds[cur.inx];
    mbSetObj$analSet$cur.inx <- cur.inx;
  }
  
  set.seed(rn.sd);
  # save the seed
  mbSetObj$analSet$rn.seed <- rn.sd;
  
  if(mbSetObj$module.type=="sdp"){
    taxrank<-"OTU";
    data <- mbSetObj$dataSet$norm.phyobj;
    data1 <- as.data.frame(t(otu_table(data)));
  }else{
    if(!exists("phyloseq_objs")){
      phyloseq_objs <- readRDS("phyloseq_objs.RDS")
    }
    
    if(taxrank=="OTU"){
      data1 <- t(phyloseq_objs$count_tables$OTU)
    }else{
      taxrank.inx <- which(names(phyloseq_objs$count_tables) %in% taxrank)
      data1 <- t(phyloseq_objs$count_tables[[taxrank.inx]])
    } 
  }
  
  data.impfeat <<- data1;
  cls <- sample_data(mbSetObj$dataSet$norm.phyobj)[[variable]];
  variable <<- variable;
  rf_out <- randomForest(data1,cls, ntree = treeNum, mtry = tryNum, importance = TRUE, proximity = TRUE);
  # set up named sig table for display
  impmat <- rf_out$importance;
  impmat <- impmat[rev(order(impmat[,"MeanDecreaseAccuracy"])),]
  sigmat <- impmat[,"MeanDecreaseAccuracy", drop=F];
  sigmat <- signif(sigmat, 5);
  write.csv(sigmat,file="randomforests_sigfeatures.csv");
  
  mbSetObj$analSet$cls <- cls;
  mbSetObj$analSet$rf <- rf_out;
  mbSetObj$analSet$rf.sigmat <- sigmat;
  return(.set.mbSetObj(mbSetObj))
}

#'@export
PlotRF.Classify<-function(mbSetObj, feature, imgName, format="png", dpi=72, width=NA){
  
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  imgName = paste(imgName,".", format, sep="");
  
  mbSetObj$imgSet$rf.cls <- imgName;
  
  if(is.na(width)){
    if(feature < 5 ){
      h <- feature*1.2;
      w <- 9;
    } else if (feature < 10){
      h <- feature*1.4;
      w <- 9;
    } else if (feature < 15){
      h <- feature/1.6;
      w <- 9;
    } else if (feature < 20){
      h <- feature/1.8;
      w <- 9;
    } else if (feature < 25){
      h <- feature/2;
      w <- 9;
    } else if (feature < 30){
      h <- feature/2.2;
      w <- 9;
    } else if (feature < 40){
      h <- feature/2.5;
      w <- 9;
    } else {
      h <- feature/10;
      w <- 9;
    }
  }else if(width == 0){
    w <- 8;
  }else{
    w <- width;
  }
  
  Cairo::Cairo(file = imgName, unit="in", dpi=dpi, width=w, height=h, type=format, bg="white");
  par(mar=c(4,4,3,2));
  cols <- grDevices::rainbow(length(levels(mbSetObj$analSet$cls))+1);
  plot(mbSetObj$analSet$rf, main="Random Forest Classification", col=cols);
  legend("topright", legend = c("Overall", levels(mbSetObj$analSet$cls)), lty=2, lwd=1, col=cols);
  dev.off();
  return(.set.mbSetObj(mbSetObj))
}

#'@export
PlotRF.VIP<-function(mbSetObj, feature, imgName, format="png", dpi=72, width=NA){
  
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  imgName = paste(imgName, ".", format, sep="");
  mbSetObj$imgSet$rf.imp <- imgName;
  vip.score <- rev(sort(mbSetObj$analSet$rf$importance[,"MeanDecreaseAccuracy"]));
  cls <- sample_data(mbSetObj$dataSet$norm.phyobj)[[variable]];
  cls.length <- length(levels(cls))
  cls.width <- cls.length/2
  
  if(is.na(width)){
    if(feature < 5 ){
      h <- feature*1.2;
    } else if (feature < 10){
      h <- feature*1.4;
    } else if (feature < 15){
      h <- feature/1.6;
    } else if (feature < 20){
      h <- feature/1.8;
    } else if (feature < 25){
      h <- feature/2;
    } else if (feature < 30){
      h <- feature/2.2;
    } else if (feature < 40){
      h <- feature/2.5;
    } else {
      h <- feature/10;
    }
    
    if(cls.length < 5){
      w <- 9.25;
    } else if(cls.length < 10){
      w <- 11.5;
    } else if(cls.length < 15){
      w <- 12.5;
    }else{
      w <- 15;
    }
  }else if(width == 0){
    w <- 8;
  }else{
    w <- width;
  }
  
  Cairo::Cairo(file = imgName,  unit="in", dpi=dpi, width=w, height=h, type=format, bg="white");
  PlotImpVar(mbSetObj, vip.score,"MeanDecreaseAccuracy", feature);
  dev.off();
  return(.set.mbSetObj(mbSetObj))
}

PlotImpVar <- function(mbSetObj, imp.vec, xlbl, feature, color.BW=FALSE){
  
  mbSetObj <- .get.mbSetObj(mbSetObj);
  
  cls.len <- length(levels(mbSetObj$analSet$cls));
  
  if(cls.len == 2){
    rt.mrg <- 5;
  }else if(cls.len == 3){
    rt.mrg <- 6;
  }else if(cls.len == 4){
    rt.mrg <- 7;
  }else if(cls.len == 5){
    rt.mrg <- 8;
  }else if(cls.len == 6){
    rt.mrg <- 9;
  }else{
    rt.mrg <- 11;
  }
  op <- par(mar=c(5,9,2,rt.mrg)); # set right side margin with the number of class
  
  feat.num = feature;
  if(feat.num <= 0){
    feat.num = 15;
  }
  
  if(feat.num > length(imp.vec)){
    feat.num <- length(imp.vec);
  }
  
  # first get the top subset
  imp.vec <- rev(sort(imp.vec))[1:feat.num];
  
  # reverser the order for display
  imp.vec <- sort(imp.vec);
  
  # as data should already be normalized, use mean/median should be the same
  # mns is a list contains means of all vars at each level
  # conver the list into a matrix with each row contains var averages across different lvls
  data1 <- data.impfeat;
  mns <- by(data1[, names(imp.vec)], mbSetObj$analSet$cls,
            function(x){ # inner function note, by send a subset of dataframe
              apply(x, 2, mean, trim=0.1)
            });
  mns <- t(matrix(unlist(mns), ncol=feat.num, byrow=TRUE));
  
  vip.nms <- names(imp.vec);
  names(imp.vec) <- NULL;
  
  # modified for B/W color
  dotcolor <- ifelse(color.BW, "darkgrey", "#585855");
  dotchart(imp.vec, bg=dotcolor, xlab= xlbl, cex=1.35);
  
  mtext(side=2, at=1:feat.num, vip.nms, las=2, line=1, cex=1.1)
  
  axis.lims <- par("usr"); # x1, x2, y1 ,y2
  
  # get character width
  shift <- 2*par("cxy")[1];
  lgd.x <- axis.lims[2] + shift;
  
  x <- rep(lgd.x, feat.num);
  y <- 1:feat.num;
  par(xpd=T);
  
  load_rcolorbrewer();
  
  nc <- ncol(mns);
  
  # modified for B/W color
  colorpalette <- ifelse(color.BW, "Greys", "RdYlBu");
  col <- grDevices::colorRampPalette(RColorBrewer::brewer.pal(10, colorpalette))(nc); # set colors for each class
  if(color.BW) col <- rev(col);
  
  # calculate background
  bg <- matrix("", nrow(mns), nc);
  for (m in 1:nrow(mns)){
    bg[m,] <- (col[nc:1])[rank(mns[m,])];
  }
  
  cls.lbl <- levels(mbSetObj$analSet$cls);
  
  for (n in 1:ncol(mns)){
    points(x,y, bty="n", pch=22, bg=bg[,n], cex=3);
    # now add label
    text(x[1], axis.lims[4], cls.lbl[n], srt=45, adj=c(0.2,0.5), cex=1.1);
    # shift x, note, this is good for current size
    x <- x + shift/1.25;
  }
  
  # now add color key, padding with more intermediate colors for contiuous band
  col <- grDevices::colorRampPalette(RColorBrewer::brewer.pal(25, colorpalette))(50)
  
  if(color.BW) col <- rev(col);
  
  nc <- length(col);
  x <- rep(x[1] + shift, nc);
  
  shifty <- (axis.lims[4]-axis.lims[3])/3;
  starty <- axis.lims[3] + shifty;
  endy <- axis.lims[3] + 2*shifty;
  y <- seq(from = starty, to = endy, length = nc);
  
  points(x,y, bty="n", pch=15, col=rev(col), cex=2);
  
  text(x[1], endy+shifty/8, "High", cex=1.1);
  text(x[1], starty-shifty/8, "Low", cex=1.1);
  par(op);
}

GetRandomNumbers <- function(){
  rm(.Random.seed);
  runif(1);
  return(123456);
}
###################################

